"""
This code is distributed under GPL license
"""

import random
import matplotlib
matplotlib.use("GTK3Agg")
import matplotlib.pyplot as plt

from tkinter import *
from tkinter.ttk import Frame, Label, Style
from tkinter import messagebox

from dataclasses import dataclass


# Graphics params
SCREEN_RESOLUTION = "400x600"
L_PAD = 15
R_PAD = 10
T_PAD = 15

# Simulation paramaters
params = {
    "a": 2,   # Alfabeto (binario) # TODO: Per permettere a!=2 devi modificare dove lo inverto
    "n": 5,   # N. righe
    "k": 2,   # Interdipendenze nella stessa riga
    "c": 3,   # Interdipendenze in altre colonne
    "x": 2,   # Quante altre colonne
    "s": 5,   # N colonne

    "steps": 300, # Quanti cicli fare su tutta la matrice
    "mutation_start": 20, # Dopo quanti step inizia a mutare
    "mutation_step": 10,  # Ogni quanti step effettuare una mutazione
    "mutation_freeze": 100, # Per quanti step freezare un allele mutato
    "kill_start": 20, # Dopo quanti step inizia le estinzioni
    "kill_step": 10,  # Ogni quanti step estinguere
    "threshold": 10, # Killa tutte le specie con fitness < di avg_fitness - threshold%
}

# Error codes
ERRS_CODE = {
    100: "'a'!=2 not supported yet",
    101: "'n' must be > 0",
    102: "'k' must be > 0",
    103: "'c' must be >= 0",
    104: "'x' must be >= 0",
    105: "'s' must be > 0",
    106: "'steps' must be > 0",
    107: "'mutation_start' must be >= 0",
    108: "'mutation_step' must be >= 0",
    109: "'kill_start' must be >= 0",
    110: "'kill_step' must be >= 0",
    111: "'threshold' must be >= 0 and <= 100",
    112: "'mutation_freeze' must be >= 0",

    200: "'k' must be < of 'n'.",
    201: "'c' must be >= of 'x'.",
    202: "'s' must be >= of 'x'.",
}


class Elem:
    """Representation of an allele
    """
    def __init__(self, val, neig):
        self.val = val              # Random [0,1] se a = 2
        self.neighborhood = neig    # Colonna -> Riga
        self.fitness_matrix = {}    # Colonna -> # fitness_matrix{"1000"} -> [0.5, 0.7]
        self.fitness = 0            # Fitness corrente
        self.mutated = 0            # Mutazione


# Here, we are creating our class, Window, and inheriting from the Frame
# class. Frame is a class from the tkinter module. (see Lib/tkinter/__init__)
class Window(Frame):
    
    def __init__(self, master=None):
        # Parameters that you want to send through the Frame class. 
        Frame.__init__(self, master)   

        # Reference to the master widget, which is the tk window                 
        self.master = master
        self.entries = {}
    

    def init_window(self):
        # https://www.python-course.eu/tkinter_entry_widgets.php
        # Changing the title of our master widget      
        self.master.title("NKCS Model")
    
        row = 0

        for k, v in params.items():
            Label(self.master, text=k+":").grid(row=row, column=0, padx=(L_PAD, R_PAD))

            e = Entry(self.master)
            e.grid(row=row, column=1)
            e.insert(10, v)
            self.entries[k] = e

            row += 1

        # Creating a button instance
        start_btn = Button(self.master, text="Start", command=(lambda e=self.entries: logic(e))).grid(row=20, column=1, padx=(L_PAD, R_PAD), pady=(T_PAD,0))


def random_list(l, n, e):
    """
    Genera n sample random in un range da 0 a L facendo
    in modo che non esca e
    l = len lista
    n = n samples
    e = elem da escludere
    """
    r_list = random.sample(range(l), n+1)
    index = n
    for i, r in enumerate(r_list):
        if r == e:
            index = i
    r_list.pop(index)
    return r_list
            

def random_exclude(n, e):
    """
    Ritorna un random int da 0 a n escluso che non sia e
    """
    r = e
    while r == e:
        r = random.randint(0, n)
    
    return r
            

def random_fitness(a):
    """
    Ritorna una lista lunga a di valori random
    fra 0 e 1
    
    Arguments:
        a {int} -- lunghezza lista
    """
    f = []
    for i in range(a):
        f.append(random.random())
    return f


def init_neighborhood(p, c, r):
    """
    p = params
    c = colonna dell'elemento
    r = riga dell'elemento
    """
    neig = []
    # Genero le interdipendenze nella stessa colonna
    r_list = random_list(p["n"], p["k"], r)

    for y in r_list:
        neig.append((c, y))

    # Genero le interdipendenze con le altre colonne
    # Nota: ogni x colonna deve avere almeno un gene
    # collegato, fino a un massimo di c geni collegati
    # fra tutte le colonne
    cols = random_list(p["s"], p["x"], c)

    for col in cols:
        row = random.randint(0, p["n"]-1)
        neig.append((col, row))

    # Costruisco un urna con i vicini possibili, cioè
    # quelli che appartengono alle stesse x colonne di
    # sopra
    urna = []
    for i in cols:
        for j in range(0, p["n"]):
            urna.append((i, j))
    
    # Tolgo dall'urna i vicini che sono già in neig
    urna = [u for u in urna if u not in neig]

    random.shuffle(urna)
    for i in range(0, p["c"]-p["x"]):
        neig.append(urna[i])
    
    return neig


def init_matrix(n, s):
    """
    Inizializza la matrice di elementi
    assegnando una fitness iniziale
    
    Arguments:
        n {int} -- num. righe
        s {int} -- num. colonne
    """
    # Crea una matrice vuota di s colonne, ognuna con n elem
    m = [ [ None for y in range( n ) ] for x in range( s ) ]
    
    # Popoliamo tutti gli elementi
    for j in range(s):
        for i in range(n):
            m[j][i] = Elem(random.randint(0,1), init_neighborhood(params, j, i))

    # Calcolo la fitness iniziale di ogni elemento
    for j in range(s):
        for i in range(n):
            neig = m[j][i].neighborhood # Lista dei vicini dell'elemento
            # Calcolo la chiave dei vicini es "10001"
            key = ""
            for elem in neig:
                key += str(m[elem[0]][elem[1]].val)

            # Ottengo il valore di fitness corrispondente alla key
            # Se tale valore ancora non esiste, lo creo
            f = m[j][i].fitness_matrix.get(key, random_fitness(params["a"]))
            m[j][i].fitness_matrix[key] = f
            m[j][i].fitness = f[m[j][i].val]

    return m
    

def step_for_col(n, s, m, already_killed):
    for j in range(s):

        if j in already_killed: 
            yield get_avg_fitness(s, n, m, already_killed, False)
            continue

        for i in range(n):

            # Se l'allele è mutante lo salto
            if m[j][i].mutated > 0:
                m[j][i].mutated -= 1
                continue

            # Calcolo la fitness del genotipo (colonna)
            f_genotipo = 0
            for allele in range(n):
                f_genotipo += m[j][i].fitness
            
            m[j][i].val = abs(m[j][i].val-1) # Se 0 metto 1 se 1 metto 0
            
            # Ricalcolo la nuova fitness
            f_new = 0
            for allele in range(n):
                neig = m[j][allele].neighborhood # Lista dei vicini dell'elemento

                # Calcolo la chiave dei vicini es "10001"
                key = ""
                for elem in neig:
                    key += str(m[elem[0]][elem[1]].val)

                # Estraggo o genero il valore di fitness
                f = m[j][allele].fitness_matrix.get(key, random_fitness(params["a"]))
                m[j][allele].fitness_matrix[key] = f

                f_new += f[m[j][allele].val]
            
            # Rimetto il valore di prima se la fitness nuova peggiora
            if f_new < f_genotipo:
                m[j][i].val = abs(m[j][i].val-1) # Se 0 metto 1 se 1 metto 0
            else:
                neig = m[j][i].neighborhood # Lista dei vicini dell'elemento

                # Calcolo la chiave dei vicini es "10001"
                key = ""
                for elem in neig:
                    key += str(m[elem[0]][elem[1]].val)

                # Estraggo o genero il valore di fitness
                f = m[j][i].fitness_matrix.get(key, random_fitness(params["a"]))
                m[j][i].fitness_matrix[key] = f
                m[j][i].fitness = f[m[j][i].val]

        yield get_avg_fitness(s, n, m, already_killed, False)


def mutation(m, j, i):
    """Muta l'allele i nella colonna j della matrice
    Mutare significa assegnare un altro valore.
    Il gene non può cambiare stato per tanti step quanto
    indicato nei paramentri
    
    Arguments:
        m {matrix[][]} -- matrice con tutte le popolazioni
        j {int} -- indice della colonna
        i {int} -- indice dell'allele
    
    Keyword Arguments:
        step {int} -- Per quanti step rimane immutato (default: {100})
    """
    m[j][i].val = abs(m[j][i].val-1) # Se 0 metto 1 se 1 metto 0
    neig = m[j][i].neighborhood # Lista dei vicini dell'elemento
    m[j][i].mutated = params["mutation_freeze"]

    # Calcolo la chiave dei vicini es "10001"
    key = ""
    for elem in neig:
        key += str(m[elem[0]][elem[1]].val)

    # Estraggo o genero il valore di fitness
    f = m[j][i].fitness_matrix.get(key, random_fitness(params["a"]))
    m[j][i].fitness_matrix[key] = f
    m[j][i].fitness = f[m[j][i].val]


def update_neighborhood(m, s, n, killed):
    if len(killed) <= 0: return

    for j in range(s):
        for i in range(n):
            # print(killed)
            # print(m[j][i].neighborhood)
            out_neigh = [i for i in m[j][i].neighborhood if i[0] not in killed]
            m[j][i].neighborhood = out_neigh
            # print(m[j][i].neighborhood)
            # print("\n\n")
            # input()


def kill_agents(m, s, n, avg_f, already_killed):
    """Kill all the species with fitness bellow the avg_fitness
    by a certain percentage.
    
    Arguments:
        m {[type]} -- [description]
        s {[type]} -- [description]
        n {[type]} -- [description]
        avg_f {[type]} -- [description]
        already_killed {[type]} -- [description]
    
    Returns:
        [list] -- List of index of the cols of the species killed.
    """
    thr = avg_f / 100 * (100-params["threshold"])

    killed = []
    for j in range(s):
        if j in already_killed: continue

        # Se la fitness di un agente è < della
        # fitness media - %threshold
        f = get_avg_col_fitness(m, j, n)
        if f < thr:
            print("Killo {} che ha {}".format(j, f))
            killed.append(j)
    
    update_neighborhood(m, s, n, killed)
    
    return killed


def get_avg_col_fitness(m, j, n):
    f = 0
        
    for i in range(n):
        f += m[j][i].fitness
        
    return f/n


def get_avg_fitness(s, n, m, killed, all=True):
    """
    Calcola la fitness media di ogni colonna
    e la media delle fitness totali
    """    
    f_list = []
    for j in range(s):   
        if j not in killed: 
            f_list.append(get_avg_col_fitness(m,j,n))

    if all: return f_list, sum(f_list)/s
    return sum(f_list)/(s-len(killed))


def print_col_fitness(s, n, m, killed):
    """Debug function.
    Print for each species the avg fitness or X if the species is dead.
    
    Arguments:
        s {int} -- Number of cols (species).
        n {int} -- Number of row (alleli).
        m {matrix[][]} --  Matrix of population n*s.
        killed {list} -- List of index of the columns of the species killed.
    """
    f_list = []
    for j in range(s):   
        if j not in killed: 
            print(get_avg_col_fitness(m,j,n), end=" ")
        else:
            print("X", end=" ")
    print("")


def update(species, plot_col, plot_epoch, mutation_list, killed_list, killed_time):
    """One step of update of the simulation

    Arguments:
        species {matrix[][]} -- Matrix of population n*s.
        plot_col {list} -- Fitness value after every column update, used to plot.
        plot_epoch {list} -- Fitness value after every step update, used to plot.
        mutation_list {list} -- List of steps in which mutations occurred, used to plot.
        killed_list {list} -- List of index of columns of species killed.
        killed_time {list} -- List of step in which some species were killed, used to plot.
    """
    for i in range(params["steps"]):

        # print_col_fitness(params["s"], params["n"], species, killed_list)

        # Mutazione : estraggo un allele a caso
        if i > params["mutation_start"] and i % params["mutation_step"] == 0:
            im = random.randint(0, params["n"]-1)
            jm = random.randint(0, params["s"]-1)
            mutation_list.append(i)
            mutation(species, jm, im)
            print("Mutazione epoca {}, gene[{}-{}]".format(i, jm, im))

        # Fai un giro di step con update per colonna
        for j in step_for_col(params["n"], params["s"], species, killed_list):
            plot_col.append(j)

        # Avg fitness
        avg_f = get_avg_fitness(params["s"], params["n"], species, killed_list, False)
        plot_epoch.append(avg_f)

        # Killo gli agenti che non fittano da una certa epoca in poi
        if i > params["kill_start"] and i % params["kill_step"] == 0:
            killed = kill_agents(species, params["s"], params["n"], avg_f, killed_list)
            
            if len(killed) > 0:
                killed_time.append(i)
                killed_list = list(set(killed_list + killed))
            
            if len(killed_list) > 0:
                print("killati: ", killed_list)


def plot(plot_col, plot_epoch, mutation_list, killed_time):
   
    # TODO: fai che puoi scegliere quali stampare
    plt.figure(1)
    for i in mutation_list:
        plt.axvline(x=i*params["s"], color="green")
    for i in killed_time:
        plt.axvline(x=i*params["s"], color="red")
    plt.plot(plot_col)
   
    plt.figure(2)
    for i in mutation_list:
        plt.axvline(x=i, color="green")
    for i in killed_time:
        plt.axvline(x=i, color="red")
    plt.plot(plot_epoch)
    
    plt.show()


def update_params(e):
    """Update params from GUI input
    
    Arguments:
        e {dict} -- params from GUI input
    """
    # Update params
    if type(e) == dict and len(e) > 0:
        for k,v in params.items():
            params[k] = int(e[k].get())


def check_params():
    """Check if param in params are correct.
    
    Returns:
        {int} -- 0 if no error, else error code. See ERRS_CODE.
    """
    # Invalid ranges
    if params["a"] != 2:
        return 100
    if params["n"] < 1:
        return 101
    if params["k"] < 0:
        return 102
    if params["c"] < 0:
        return 103
    if params["x"] < 0:
        return 104
    if params["s"] < 1:
        return 105
    if params["steps"] < 1:
        return 106
    if params["mutation_start"] < 0:
        return 107
    if params["mutation_step"] < 0:
        return 108
    if params["kill_start"] < 0:
        return 109
    if params["kill_step"] < 0:
        return 110
    if params["threshold"] < 0 or params["threshold"] > 100:
        return 111
    if params["mutation_freeze"] < 0:
        return 112
        

    # Inconsistencies
    if params["k"] > params["n"]:
        return 200
    if params["c"] < params["x"]:
        return 201
    if params["s"] < params["x"]:
        return 202

    return 0


def logic(e):
    """Init the logic data and run the simulation
    
    Arguments:
        e {dict} -- Input params from GUI. See params.
    """
    # Update params from input and check if there are no errors
    update_params(e)
    err = check_params()
    if err > 0:
        messagebox.showerror("Error", ERRS_CODE.get(err, "Generic error!"))
        return

    # Data
    species = init_matrix(params["n"], params["s"])
    plot_col = []
    plot_epoch = []
    mutation_list = []
    killed_list = []
    killed_time = []

    # Step loop
    update(species, plot_col, plot_epoch, mutation_list, killed_list, killed_time)
    
    # Plot
    plot(plot_col, plot_epoch, mutation_list, killed_time)


if __name__ == "__main__":

    root = Tk()
    root.geometry(SCREEN_RESOLUTION)
    app = Window(root)
    app.init_window()
    root.mainloop()
